/*
Copyright (C) 2001 Lkb

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

enum MuxValue
{
	MUX_0 = 0,
	MUX_1 = 1,
	MUX_COMBINED = 2,
	MUX_TEXEL0 = 3,
	MUX_TEXEL1 = 4,
	MUX_PRIM = 5,
	MUX_SHADE = 6,
	MUX_ENV = 7,
	MUX_COMBALPHA = 8,
	MUX_T0_ALPHA = 9,
	MUX_T1_ALPHA = 10,
	MUX_PRIM_ALPHA = 11,
	MUX_SHADE_ALPHA = 12,
	MUX_ENV_ALPHA = 13,
	MUX_LODFRAC = 14,
	MUX_PRIMLODFRAC = 15,
	MUX_UNK = 16
};

extern const MuxValue g_Mux32[32];
extern const MuxValue g_Mux16[16];
extern const MuxValue g_Mux8[8];

extern vector<const char*> g_combinerErrorMessages;

extern const char * g_colcombtypes32[32];
extern const char * g_colcombtypes16[16];
extern const char * g_colcombtypes8[8];

extern const char* g_variableStrings[];
extern const char* g_textureOpStrings[];
extern const char* g_textureOpRawStrings[];
extern const char* g_textureArgumentStrings[];

extern const char* g_d3dblendStrings[];
extern const char* g_constantSlotStrings[];

struct CombinerColorComponent
{
	typedef float float_type;
	static const float_type f0;
	static const float_type f1;
	static const float_type f2;
	static const float_type f0_5;
	static const float_type f255;
    float_type value;

    CombinerColorComponent()
		//: value(f0)
	{}

    CombinerColorComponent(float_type v)
		: value(v)
	{}

	CombinerColorComponent(u8 v)
		: value(((float_type)v) / f255)
	{}

	CombinerColorComponent(int v)
		: value(((float_type)v) / f255)
	{}

	CombinerColorComponent(unsigned int v)
		: value(((float_type)v) / f255)
	{}

	operator u8() const
	{
		return (u8)((f255 * value) + f0_5);
	}
    
	operator unsigned int() const
	{
		return (unsigned int)((f255 * value ) + f0_5);
	}
    
	operator float_type() const
	{
		return value;
	}

    friend CombinerColorComponent operator+(const CombinerColorComponent& a, const CombinerColorComponent& b)
    {
        return CombinerColorComponent(min(a.value + b.value, f1));
    }

    friend CombinerColorComponent operator -(const CombinerColorComponent& a, const CombinerColorComponent& b)
    {
        return CombinerColorComponent(max(a.value - b.value, f0));
    }

    friend CombinerColorComponent operator*(const CombinerColorComponent& a, const CombinerColorComponent& b)
    {
        return CombinerColorComponent(a.value * b.value);
    }

	CombinerColorComponent operator ~() const
	{
		return CombinerColorComponent(f1 - value);
	}

	CombinerColorComponent& operator +=(CombinerColorComponent v) {*this = *this + v; return *this;}
	CombinerColorComponent& operator -=(CombinerColorComponent v) {*this = *this - v; return *this;}
	CombinerColorComponent& operator *=(CombinerColorComponent v) {*this = *this * v; return *this;}
};

struct CombinerColor
{
	CombinerColorComponent r, g, b, a;

	CombinerColor()
		//: r(), g(), b(), a()
	{

	}
	
	CombinerColor(const CombinerColorComponent& p_r, const CombinerColorComponent& p_g, const CombinerColorComponent& p_b, const CombinerColorComponent& p_a)
		: r(p_r), g(p_g), b(p_b), a(p_a)
	{}

	CombinerColor(const CombinerColorComponent& v)
		: r(v), g(v), b(v), a(v)
	{}

	CombinerColor(CombinerColorComponent::float_type v)
		: r(v), g(v), b(v), a(v)
	{}

	CombinerColor(u8 v)
		: r(v), g(v), b(v), a(v)
	{}

	CombinerColor(int v)
		: r(v), g(v), b(v), a(v)
	{}

	CombinerColor(unsigned int v)
		: r(v), g(v), b(v), a(v)
	{}

	CombinerColor(D3DCOLOR v)
		: r((u8)RGBA_GETRED(v)), g((u8)RGBA_GETGREEN(v)), b((u8)RGBA_GETBLUE(v)), a((u8)RGBA_GETALPHA(v))
	{}

	friend CombinerColor operator +(const CombinerColor& left, const CombinerColor& right)
	{
		CombinerColor color;
		color.r = left.r + right.r;
		color.g = left.g + right.g;
		color.b = left.b + right.b;
		color.a = left.a + right.a;
		return color;
	}

	friend CombinerColor operator -(const CombinerColor& left, const CombinerColor& right)
	{
		CombinerColor color;
		color.r = left.r * right.r;
		color.g = left.g * right.g;
		color.b = left.b * right.b;
		color.a = left.a * right.a;
		return color;
	}

	friend CombinerColor operator *(const CombinerColor& left, const CombinerColor& right)
	{
		CombinerColor color;
		color.r = left.r * right.r;
		color.g = left.g * right.g;
		color.b = left.b * right.b;
		color.a = left.a * right.a;
		return color;
	}

	CombinerColor AlphaReplicate() const
	{
		CombinerColor color;
		color.r = a;
		color.g = a;
		color.b = a;
		return color;
	}

	CombinerColor operator ~() const
	{
		return CombinerColor(1) - *this;
	}

	friend CombinerColor MergeColorAndAlpha(const CombinerColor& left, const CombinerColor& right)
	{
		return CombinerColor(left.r, left.g, left.b, right.a);
	}

	friend CombinerColor MergeColorAndAlpha(const CombinerColor& left, const CombinerColorComponent alpha)
	{
		return CombinerColor(left.r, left.g, left.b, alpha);
	}

	CombinerColor& operator +=(const CombinerColor& v) {*this = *this + v; return *this;}
	CombinerColor& operator -=(const CombinerColor& v) {*this = *this - v; return *this;}
	CombinerColor& operator *=(const CombinerColor& v) {*this = *this * v; return *this;}
	
	/*
	CombinerColor& operator =(const CombinerColorComponent& v)
	{
		r = v;
		g = v;
		b = v;
		a = v;
		return *this;
	}

	CombinerColor& operator =(CombinerColorComponent::float_type v)	{return (*this = CombinerColorComponent(v));}
	CombinerColor& operator =(u8 v)									{return (*this = CombinerColorComponent(v));}
	CombinerColor& operator =(int v)								{return (*this = CombinerColorComponent(v));}
	CombinerColor& operator =(unsigned int v)						{return (*this = CombinerColorComponent(v));}
	*/

	operator D3DCOLOR() const
	{
		return RGBA_MAKE((u8)r, (u8)g, (u8)b, (u8)a);
	}
};

enum CombinerValueFlags
{
	CombinerValueFlags_Variable = 1,
	CombinerValueFlags_Expression = 2,

	CombinerValueFlags_TypeMask = 0xf,

	CombinerValueFlags_Complement = 0x10,
	CombinerValueFlags_GetAlpha = 0x20,
	CombinerValueFlags_InAlpha = 0x40,

	CombinerValueFlags_Known = 0x100,
	CombinerValueFlags_KnownExpression = 0x102,
	CombinerValueFlags_MultiPass = 0x200,

	CombinerValueFlags_EqualityMask = 0x3f,

	CombinerValueFlags_ConstantSlotEqualityMask = 0x2f,
};

enum CombinerVariable
{
	// same values as corresponding D3DTA_
	CombinerVariable_Diffuse = 0,
	CombinerVariable_Current = 1,
	CombinerVariable_Texture = 2,
	CombinerVariable_TFactor = 3,
	CombinerVariable_Specular = 4,

	CombinerVariable_PreviousPass = 5,
	
	CombinerVariable_Unused = 6, 

	CombinerVariable_0 = 8,
	CombinerVariable_Primitive = 9,
	CombinerVariable_Environment = 10,

	CombinerVariable_FirstConstant = 8,
	//CombinerVariable_Unknown = 8, // TODO: handle this better

	CombinerVariable_Nop = 11
};

struct CombinerValue;
struct CombinerExpression;

vector<CombinerExpression*> allocatedExpressions;

struct CombinerExpression
{
	D3DTEXTUREOP operation;
	vector<CombinerValue> values;

	CombinerExpression(D3DTEXTUREOP p_operation, CombinerValue p_left, CombinerValue p_right);
	CombinerExpression::CombinerExpression(D3DTEXTUREOP p_operation, CombinerValue p_operand);

	CombinerExpression(D3DTEXTUREOP p_operation = D3DTOP_DISABLE, vector<CombinerValue> p_values = vector<CombinerValue>())
		: operation(p_operation), values(p_values)
	{}

	friend bool operator ==(const CombinerExpression& v1, const CombinerExpression& v2);

	friend bool operator !=(const CombinerExpression& v1, const CombinerExpression& v2)
	{
		return !(v1 == v2);
	}

	friend bool operator <(const CombinerExpression& v1, const CombinerExpression& v2);

	void MarkAsInUse(bool mark)
	{
		if(mark)
		{
			operation = (D3DTEXTUREOP)(operation | 0x80);
		}
		else
		{
			operation = (D3DTEXTUREOP)(operation &~ 0x80);
		}
	}

	bool IsInUse()
	{
		return (operation & 0x80) ? true : false;
	}

	void *operator new(size_t size)
	{
		void* p = malloc(size);
		allocatedExpressions.push_back((CombinerExpression*)p);
		return p;
	}

	// delete must be called only by the garbage collector!
};

struct CombinerPreliminarStage;
struct CombinerStage;
struct CombinerPreliminarPass;
struct CombinerPass;

#ifdef _DEBUG
#define CombinerValue_InitCheckingOperators checkingOperators(false), 
#else
#define CombinerValue_InitCheckingOperators
#endif

struct CombinerValue
{
	union
	{
		DWORD value;
		CombinerVariable variable;
		CombinerExpression* expression;
	};
	CombinerValueFlags flags;
	BYTE depth;
	
	CombinerValue(bool p_getAlpha = false, bool p_complement = false)
		: CombinerValue_InitCheckingOperators flags((CombinerValueFlags)(0 | (p_getAlpha ? CombinerValueFlags_GetAlpha : 0) | (p_complement ? CombinerValueFlags_Complement : 0))), value(0)
	{}

	CombinerValue(CombinerVariable p_variable, bool p_getAlpha = false, bool p_complement = false)
		: CombinerValue_InitCheckingOperators flags((CombinerValueFlags)(CombinerValueFlags_Variable | (p_getAlpha ? CombinerValueFlags_GetAlpha : 0) | (p_complement ? CombinerValueFlags_Complement : 0))), variable(p_variable)
	{}

	CombinerValue(CombinerExpression* p_expression, bool p_getAlpha = false, bool p_complement = false)
		: CombinerValue_InitCheckingOperators flags((CombinerValueFlags)(CombinerValueFlags_Expression | (p_getAlpha ? CombinerValueFlags_GetAlpha : 0) | (p_complement ? CombinerValueFlags_Complement : 0))), expression(p_expression)
	{}

	CombinerValue(D3DTEXTUREOP p_operation, bool p_getAlpha = false, bool p_complement = false)
		: CombinerValue_InitCheckingOperators flags((CombinerValueFlags)(CombinerValueFlags_Expression | (p_getAlpha ? CombinerValueFlags_GetAlpha : 0) | (p_complement ? CombinerValueFlags_Complement : 0))), expression(new CombinerExpression(p_operation))
	{}

	CombinerValue(D3DTEXTUREOP operation, CombinerValue left, CombinerValue right, bool p_getAlpha = false, bool p_complement = false)
	 : CombinerValue_InitCheckingOperators flags((CombinerValueFlags)(CombinerValueFlags_Expression | (p_getAlpha ? CombinerValueFlags_GetAlpha : 0) | (p_complement ? CombinerValueFlags_Complement : 0))), expression(new CombinerExpression(operation, left, right))
	{}

	CombinerValue(D3DTEXTUREOP operation, CombinerValue operand, bool p_getAlpha = false, bool p_complement = false)
	 : CombinerValue_InitCheckingOperators flags((CombinerValueFlags)(CombinerValueFlags_Expression | (p_getAlpha ? CombinerValueFlags_GetAlpha : 0) | (p_complement ? CombinerValueFlags_Complement : 0))), expression(new CombinerExpression(operation, operand))
	{}

	bool IsExpression() const
	{
		return (flags & CombinerValueFlags_Expression) ? true : false;
	}

	bool IsExpression(D3DTEXTUREOP top) const
	{
		return IsExpression() && (expression->operation == top);
	}

	bool IsUnknownExpression() const
	{
		return IsExpression() && !(flags & CombinerValueFlags_Known);
	}

	bool IsUnknownExpression(D3DTEXTUREOP top) const
	{
		return IsUnknownExpression() && (expression->operation == top);
	}

	bool IsKnownExpression() const
	{
		return IsExpression() && (flags & CombinerValueFlags_Known);
	}

	bool IsVariable() const
	{
		return (flags & CombinerValueFlags_Variable) ? true : false;
	}

	bool IsVariable(CombinerVariable p_variable) const
	{
		return ((flags & CombinerValueFlags_Variable) && (variable == p_variable));
	}

	bool IsUnknownVariable() const
	{
		return IsVariable() && (variable < CombinerVariable_FirstConstant);
	}

	bool IsKnownVariable() const
	{
		return IsVariable() && (variable >= CombinerVariable_FirstConstant);
		//return (flags & CombinerValueFlags_Known) || ((flags & CombinerValueFlags_Variable) && ((variable >= CombinerVariable_FirstConstant)));
	}

	bool IsKnown() const
	{
		return IsKnownVariable() || IsKnownExpression();
	}

	bool IsUnknown() const
	{
		return IsUnknownVariable() || IsUnknownExpression();
	}

	bool IsInAlpha() const
	{
		return (flags & CombinerValueFlags_InAlpha) ? true : false;
	}

	bool IsGetAlpha() const
	{
		return (flags & CombinerValueFlags_GetAlpha) ? true : false;
	}

	bool IsComplement() const
	{
		return (flags & CombinerValueFlags_Complement) ? true : false;
	}

	CombinerValue& Complement()
	{
		flags = (CombinerValueFlags)(flags ^ CombinerValueFlags_Complement);
		return *this;
	}

	CombinerValue& SetGetAlpha()
	{
		flags = (CombinerValueFlags)(flags | CombinerValueFlags_GetAlpha);
		return *this;
	}

	CombinerValue& SetInAlpha()
	{
		flags = (CombinerValueFlags)(flags | CombinerValueFlags_GetAlpha | CombinerValueFlags_InAlpha);
		return *this;
	}

	CombinerValue& SetKnown()
	{
		flags = (CombinerValueFlags)(flags | CombinerValueFlags_Known);
		return *this;
	}

	static bool CombinerValue_depth_greater(const CombinerValue& x, const CombinerValue& y)
	{
		return x.depth > y.depth;
    }

	static bool CombinerValue_depth_less(const CombinerValue& x, const CombinerValue& y)
	{
		return x.depth < y.depth;
    }

	static void MarkAsInAlpha(vector<CombinerPreliminarStage>& stages);
	DWORD ComputeDepth();
// BEGIN MODIFIED BY Lkb - 12/jul/2001 - GCC support
	CombinerValue GetValueAndXorFlags(CombinerValueFlags vand, CombinerValueFlags vxor = (CombinerValueFlags)0) const;
// END MODIFIED BY Lkb - 12/jul/2001 - GCC support
	void RearrangeBasedOnDepth();
	void OptimizeConstants();
	void RemoveNops();
	void AdjustSingleOperandExpression();
	void RemoveRedundantOperations();
	void NormalizeAddSub();
	void NormalizeGetAlpha();
	void AssociateOperations();
	void TransformToBinary();
	void OptimizeOperations();
	bool IsMultiPass() const;
	bool MarkMultiPassNodes();
	void RearrangeTextures();
	void Process();
	BYTE ConvertToTextureArgument(bool isAlpha);
	void TransformSubtractions();
	bool GenerateCombinerStages(/*[out]*/ CombinerStage* pStages, bool isAlpha, /*[in, out]*/ CombinerStage* pAlphaStages = NULL, bool testing = false);
	static bool AssignConstantSlots(CombinerValue colorPass, CombinerValue alphaPass, CombinerValue* constantSlotAssignments, map<CombinerValue, CombinerValue>& constantSlotReplacementMap);
	static bool EvaluateConstantSlots(CombinerValue colorPass, CombinerValue alphaPass);
	bool AssignConstantSlots(CombinerValue* constantSlotAssignments, map<CombinerValue, CombinerValue>& constantSlotReplacementMap);
	bool EvaluateConstantSlots();
	bool EvaluateStages(CombinerValue* pColorPass, CombinerValue* pAlphaPass);
	static bool AssignConstantSlots(map<CombinerValue, DWORD>& constantSlotMap, CombinerValue* constantSlotAssignments, map<CombinerValue, CombinerValue>& constantSlotReplacementMaps);
	
	void GenerateCombinerInfo(CombinerInfo& info, CombinerValue* pAlphaPass, D3DTEXTUREOP operation = D3DTOP_DISABLE);
	void GenerateCombinerPass(CombinerValue* pAlphaPass, CombinerPass& pass, D3DTEXTUREOP operation, bool isFirst);

	friend bool Compare(const CombinerValue& v1, const CombinerValue& v2, CombinerValueFlags mask);

#ifdef _DEBUG
	mutable bool checkingOperators;
#endif

	static void CheckOperatorConsistency(const CombinerValue& a, const CombinerValue& b);
	friend bool operator ==(const CombinerValue& v1, const CombinerValue& v2);
	friend bool operator <(const CombinerValue& v1, const CombinerValue& v2);
	friend bool operator >=(const CombinerValue& v1, const CombinerValue& v2);
	friend bool operator !=(const CombinerValue& v1, const CombinerValue& v2);
	CombinerValue DuplicateTree() const;
	//template<typename T>
	//void CombinerValue::Dump(const T& stream) const;

	template<typename T>
	void Dump(T& stream) const
	{
		if(flags & CombinerValueFlags_Complement)
		{
			stream << "(1 - ";
		}
		if(flags & CombinerValueFlags_GetAlpha)
		{
			stream << "alpha";
			if(IsVariable())
			{
				stream << "(";
			}
		}
		if(IsVariable())
		{
			stream << g_variableStrings[variable];
		}
		else if(IsExpression())
		{
			stream << "(";
			for(vector<CombinerValue>::iterator iter = expression->values.begin(); iter < expression->values.end(); iter++)
			{
				if(iter != expression->values.begin())
				{
					stream << g_textureOpStrings[expression->operation];
				}
				iter->Dump(stream);
			}
			stream << ")";
		}
		if((flags & CombinerValueFlags_GetAlpha) && IsVariable())
		{
			stream << ")";
		}
		if(flags & CombinerValueFlags_Complement)
		{
			stream << ")";
		}
	}

	CombinerColor ComputeColor(const CombinerColor& primitive, const CombinerColor& environment);
	CombinerColorComponent ComputeAlpha(const CombinerColor& primitive, const CombinerColor& environment);

	void AddConstantSlotToMap(map<CombinerValue, DWORD>& constantSlotMap) const;
	static void AddConstantSlotsToMap(const vector<CombinerPreliminarStage>& stages, map<CombinerValue, DWORD>& constantSlotMap);
	void ApplyConstantSlot(const map<CombinerValue, CombinerValue>& constantSlotMap);
	static void ApplyConstantSlots(vector<CombinerPreliminarStage>& stages, const map<CombinerValue, CombinerValue>& constantSlotMap);

	void GenerateAllPossibleInfos_MultiPass(CombinerValue* pRootValue, CombinerValue* pAlphaValue, vector<CombinerInfo>& infoVector, vector<CombinerValue>& multiPassVector, vector<CombinerValue>& zeroDepthVector, bool usingDefaultAlpha);
	void GenerateAllPossibleInfos(CombinerValue* pRootValue, CombinerValue* pAlphaValue, vector<CombinerInfo>& infoVector, bool usingDefaultAlpha, bool multiPassReordered = false);
	void GenerateInfo(CombinerInfo& info, CombinerValue* pAlphaPass, D3DTEXTUREOP operation = D3DTOP_DISABLE);
	bool GenerateStages(/*[out]*/ vector<CombinerPreliminarStage>& stages, bool isAlpha, bool useBlendAlpha);
	void GenerateBestPass(CombinerValue* pAlphaPass, CombinerPass& outPass, D3DTEXTUREOP operation, bool isFirst);
	void GenerateAllPossiblePasses_TextureConstantSlots(vector<CombinerPass>& passes, CombinerPreliminarPass& pass, int stageNum = 0, bool doAlpha = false);
	void GenerateAllPossiblePasses_Finish(vector<CombinerPass>& passes, CombinerPreliminarPass& pass);
	void GenerateAllPossiblePasses_Shift(vector<CombinerPass>& passes, CombinerPreliminarPass& pass, int stageNum, bool doAlpha = false);
	void GenerateAllPossiblePasses_Transform(vector<CombinerPass>& passes, CombinerPreliminarPass& pass, int stageNum, bool doAlpha);
	void GenerateAllPossiblePasses(vector<CombinerPass>& passes, CombinerPreliminarPass& pass);
	void GenerateBestInfo(CombinerValue* pAlphaValue, CombinerInfo& info);
	
	static DWORD EvaluatePass(CombinerPass& pass);
	static DWORD EvaluateInfo(CombinerInfo& info);

	void MarkAsInUse(bool mark);
};

#ifdef CombinerValue_InitCheckingOperators
#undef CombinerValue_InitCheckingOperators
#endif

template<typename T, typename U>
struct pair_second_greater : binary_function<bool, T, U>
{
	bool operator()(const pair<T, U>& v1, const pair<T, U>& v2)
	{
		return v1.second > v2.second;
	}
};

struct CombinerValueWithUsageCount
{
	CombinerValue value;
	DWORD usage;

	CombinerValueWithUsageCount(CombinerValue p_value = CombinerValue(), DWORD p_usage = 0)
		: value(p_value), usage(p_usage)
	{}

	// BEGIN MODIFIED BY Lkb - 9/jun/2001 - GCC support
	friend bool operator ==(const CombinerValueWithUsageCount& v1, const CombinerValueWithUsageCount& v2)
	{
		return v1.value == v2.value;
	}

	friend bool operator !=(const CombinerValueWithUsageCount& v1, const CombinerValueWithUsageCount& v2)
	{
		return v1.value == v2.value;
	}

	friend bool operator <(const CombinerValueWithUsageCount& v1, const CombinerValueWithUsageCount& v2)
	{
		return v1.value < v2.value;
	}

	friend bool operator >=(const CombinerValueWithUsageCount& v1, const CombinerValueWithUsageCount& v2)
	{
		return v1.value >= v2.value;
	}

	static bool usage_greater(const CombinerValueWithUsageCount& v1, const CombinerValueWithUsageCount& v2)
	{
		return v1.usage > v2.usage;
	}
	// END MODIFIED BY Lkb - 9/jun/2001 - GCC support
};

enum CombinerConstantSlot
{
	CombinerConstantSlot_TFactor = 0,
	CombinerConstantSlot_Diffuse = 1,
	CombinerConstantSlot_Specular = 2,
	CombinerConstantSlot_TFactorAlpha = 3,
	CombinerConstantSlot_DiffuseAlpha = 4,
	CombinerConstantSlot_SpecularAlpha = 5,
};

struct CombinerStage
{
	D3DTEXTUREOP operation; // TODO: declare as WORD
	BYTE operand1;
	BYTE operand2;

	CombinerStage(D3DTEXTUREOP p_operation = D3DTOP_DISABLE, BYTE p_operand1 = 0, BYTE p_operand2 = 0)
		: operation(p_operation), operand1(p_operand1), operand2(p_operand2)
	{}

// BEGIN MODIFIED BY Lkb - 12/jul/2001 - GCC support
	void DumpOperand(std::ostream& stream, BYTE operand) const;
	void Dump(std::ostream& stream) const;
// END MODIFIED BY Lkb - 12/jul/2001 - GCC support
};

struct CombinerPreliminarStage
{
	D3DTEXTUREOP operation;
	CombinerValue operand1;
	CombinerValue operand2;

	CombinerPreliminarStage(D3DTEXTUREOP p_operation = D3DTOP_DISABLE, CombinerValue p_operand1 = CombinerValue(), CombinerValue p_operand2 = CombinerValue())
		: operation(p_operation), operand1(p_operand1), operand2(p_operand2)
	{}

	void ConvertToCombinerStage(CombinerStage& stage, bool isAlpha);
};

struct CombinerPass
{
	DWORD score;
	HRESULT hres;

	CombinerValue constantSlots[6];

	CombinerValue colorTextureSlots[8];
	CombinerValue alphaTextureSlots[8];

	CombinerStage colorStages[8];
	CombinerStage alphaStages[8];

	D3DBLEND srcBlend; // if 0, operation is valid
	D3DBLEND destBlend;
	
	bool alphaTest;
	bool blend;

	CombinerPass()
		: srcBlend((D3DBLEND)0), alphaTest(false), blend(false)
	{}

	static bool score_less(const CombinerPass& v1, const CombinerPass& v2)
	{
		return v1.score < v2.score;
	}
};

struct CombinerPreliminarPass
{
	CombinerValue constantSlots[6];

	HRESULT hresult;
	
	vector<CombinerValue> colorTextureSlots;
	vector<CombinerValue> alphaTextureSlots;

	vector<CombinerPreliminarStage> colorStages;
	vector<CombinerPreliminarStage> alphaStages;

	D3DBLEND srcBlend; // if 0, operation is valid
	union
	{
		D3DTEXTUREOP operation;
		D3DBLEND destBlend;
	};

	bool alphaTest;
	bool blend;

	CombinerPreliminarPass()
		: srcBlend((D3DBLEND)0), alphaTest(false), blend(false)
	{
		for(int iCSInit = 0; iCSInit < 6; iCSInit++)
		{
			constantSlots[iCSInit] = CombinerValue();
		}
	}

	void ConvertToCombinerPass(CombinerPass& pass);
};

/*
struct CombinerPreliminarPass
{
	D3DTEXTUREOP operation;
	
	CombinerValue value;

	CombinerPreliminarPass()
		: operation(D3DTOP_DISABLE)
	{}

	void GenerateCombinerPass(CombinerValue* pAlphaPass, CombinerPass& pass, bool isFirst);
};
*/

struct CombinerTempInfo
{
	u64 mux;
	DWORD incorrectness;
	HRESULT validationResult;
	vector<CombinerPass> passes;
};

// size: 
struct CombinerInfo
{
	u64 mux;
	DWORD score;
	HRESULT hres;
	vector<CombinerPass> passes;

	static bool score_less(const CombinerInfo v1, const CombinerInfo& v2)
	{
		return v1.score < v2.score;
	}

	CombinerInfo()
		: mux(0)
	{}


	// BEGIN MODIFIED BY Lkb - 12/jul/2001 - GCC support
	void DumpMux(std::ostream& stream) const
	// END MODIFIED BY Lkb - 12/jul/2001 - GCC support
	{
		DWORD dwMux0 = (DWORD)(mux>>32);
		DWORD dwMux1 = (DWORD)(mux);

		DWORD aRGB0  = (dwMux0>>20)&0x0F;	// c1 c1		// a0
		DWORD bRGB0  = (dwMux1>>28)&0x0F;	// c1 c2		// b0
		DWORD cRGB0  = (dwMux0>>15)&0x1F;	// c1 c3		// c0
		DWORD dRGB0  = (dwMux1>>15)&0x07;	// c1 c4		// d0

		DWORD aA0    = (dwMux0>>12)&0x07;	// c1 a1		// Aa0
		DWORD bA0    = (dwMux1>>12)&0x07;	// c1 a2		// Ab0
		DWORD cA0    = (dwMux0>>9 )&0x07;	// c1 a3		// Ac0
		DWORD dA0    = (dwMux1>>9 )&0x07;	// c1 a4		// Ad0

		DWORD aRGB1  = (dwMux0>>5 )&0x0F;	// c2 c1		// a1
		DWORD bRGB1  = (dwMux1>>24)&0x0F;	// c2 c2		// b1
		DWORD cRGB1  = (dwMux0    )&0x1F;	// c2 c3		// c1
		DWORD dRGB1  = (dwMux1>>6 )&0x07;	// c2 c4		// d1

		DWORD aA1    = (dwMux1>>21)&0x07;	// c2 a1		// Aa1
		DWORD bA1    = (dwMux1>>3 )&0x07;	// c2 a2		// Ab1
		DWORD cA1    = (dwMux1>>18)&0x07;	// c2 a3		// Ac1
		DWORD dA1    = (dwMux1    )&0x07;	// c2 a4		// Ad1

		TCHAR buffer[256];
		sprintf(buffer, "MUX: 0x%08x%08x\n", dwMux0, dwMux1);
		stream << buffer;
		stream << "\tColor Pass 1: (" << g_colcombtypes16[aRGB0]	<< " - " << g_colcombtypes16[bRGB0]	<< ") * " << g_colcombtypes32[cRGB0]	<< " + " << g_colcombtypes8[dRGB0]	<< "\n";
		stream << "\tAlpha Pass 1: (" << g_colcombtypes8[aA0]		<< " - " << g_colcombtypes8[bA0]		<< ") * " << g_colcombtypes8[cA0]		<< " + " << g_colcombtypes8[dA0]	<< "\n";
		stream << "\tColor Pass 2: (" << g_colcombtypes16[aRGB1]	<< " - " << g_colcombtypes16[bRGB1]	<< ") * " << g_colcombtypes32[cRGB1]	<< " + " << g_colcombtypes8[dRGB1]	<< "\n";
		stream << "\tAlpha Pass 2: (" << g_colcombtypes8[aA1]		<< " - " << g_colcombtypes8[bA1]		<< ") * " << g_colcombtypes8[cA1]		<< " + " << g_colcombtypes8[dA1]	<< "\n";
	}

	void Dump(std::ostream& stream) const
	{
		stream << "\n--------------------------------------------------------------------------------\n";
		DumpMux(stream);
		stream << "\n";
		stream << "Generated Combiner Info:";
		int i = 0;
		TCHAR buffer[256];
		for(vector<CombinerPass>::const_iterator iter = passes.begin(); iter < passes.end(); iter++, i++)
		{
			sprintf(buffer, "\nPass %u\t", i);
			stream << buffer;
			stream << "AlphaTest: " << ((iter->alphaTest) ? "On" : "Off");
			stream << " Blending: " << ((iter->blend) ? "On" : "Off");
			stream << " DestBlend: " << g_d3dblendStrings[iter->destBlend];
			stream << " SrcBlend: " << g_d3dblendStrings[iter->srcBlend];
			stream << "\n";
			
			stream << "\tConstant Slots\n";
			int i;
			for(i = 0; i < 6; i++)
			{
				if(iter->constantSlots[i].flags != 0)
				{
					stream << "\t\t" << g_constantSlotStrings[i] << ": ";
					iter->constantSlots[i].Dump(stream);
					stream << "\n";
				}
			}

			stream << "\tTextures\n";
			for(i = 0; i < 8; i++)
			{
				stream << "\t\t" << i << " - Color: ";
				iter->colorTextureSlots[i].Dump(stream);
				stream << " Alpha: ";
				iter->alphaTextureSlots[i].Dump(stream);
				stream << "\n"; 
			}

			stream << "\tColor Stages\n";
			for(i = 0; i < 8; i++)
			{
				if(iter->colorStages[i].operation == D3DTOP_DISABLE)
					break;
				stream << "\t\t";
				iter->colorStages[i].Dump(stream);
				stream << "\n";
			}

			stream << "\tAlpha Stages\n";
			for(i = 0; i < 8; i++)
			{
				if(iter->alphaStages[i].operation == D3DTOP_DISABLE)
					break;
				stream << "\t\t";
				iter->alphaStages[i].Dump(stream);
				stream << "\n";
			}
		}
		if(g_combinerErrorMessages.size() > 0)
		{
			stream << "Error Messages\n";
			for(vector<const char*>::const_iterator emiter = g_combinerErrorMessages.begin(); emiter < g_combinerErrorMessages.end(); emiter++)
			{
				stream << "\t" << *emiter << "\n";
			}
			g_combinerErrorMessages.clear();
		}
	}

	void MarkAsInUse(bool mark)
	{
		vector<CombinerPass>::iterator passIter;
		for(passIter = passes.begin(); passIter < passes.end(); passIter++)
		{
			int i;
			for(i = 0; i < 6; i++)
			{
				passIter->constantSlots[i].MarkAsInUse(mark);
			}
			for(i = 0; i < 8; i++)
			{
				passIter->colorTextureSlots[i].MarkAsInUse(mark);
				passIter->alphaTextureSlots[i].MarkAsInUse(mark);
			}
		}
	}
};
